@@ -529,14 +529,14 @@ static void show_switch_statement(struct statement *stmt)
 			} else
 				printf("    what?");
 		}
-		printf(": .L%p\n", sym->bb_target);
+		printf(": .L%p\n", sym);
 	} END_FOR_EACH_PTR(sym);
 	printf("# end case table\n");
 
 	show_statement(stmt->switch_statement);
 
 	if (stmt->switch_break->used)
-		printf(".L%p:\n", stmt->switch_break->bb_target);
+		printf(".L%p:\n", stmt->switch_break);
 }
 
 static void show_symbol_decl(struct symbol_list *syms)
@@ -683,7 +683,7 @@ int show_statement(struct statement *stmt)
 			int val = show_expression(stmt->goto_expression);
 			printf("\tgoto\t\t*v%d\n", val);
 		} else {
-			printf("\tgoto\t\t.L%p\n", stmt->goto_label->bb_target);
+			printf("\tgoto\t\t.L%p\n", stmt->goto_label);
 		}
 		break;
 	case STMT_ASM:
