@@ -20,6 +20,8 @@ extern int bits_in_long;
 extern int bits_in_longlong;
 extern int bits_in_longlonglong;
 
+extern int bits_in_wchar;
+
 extern int max_int_alignment;
 
 /*
@@ -49,7 +51,7 @@ extern int enum_alignment;
 
 static inline int bits_to_bytes(int bits)
 {
-	return bits >= 0 ? bits / bits_in_char : -1;
+	return bits >= 0 ? (bits + bits_in_char - 1) / bits_in_char : -1;
 }
 
 static inline int bytes_to_bits(int bytes)
@@ -57,4 +59,12 @@ static inline int bytes_to_bits(int bytes)
 	return bytes * bits_in_char;
 }
 
+static inline unsigned long array_element_offset(unsigned long base_bits, int idx)
+{
+	int fragment = base_bits % bits_in_char;
+	if (fragment)
+		base_bits += bits_in_char - fragment;
+	return base_bits * idx;
+}
+
 #endif
